/* $Id: Str.c,v 1.3 1999/07/29 11:50:45 okabe Exp $ */
/*
 * String manipulation library for Boehm GC
 *
 * (C) Copyright 1998-1999 by Akinori Ito
 *
 * This software may be redistributed freely for this purpose, in full 
 * or in part, provided that this entire copyright notice is included 
 * on any copies of this software and applications and derivations thereof.
 *
 * This software is provided on an "as is" basis, without warranty of any
 * kind, either expressed or implied, as to any matter including, but not
 * limited to warranty of fitness of purpose, or merchantability, or
 * results obtained from use of this software.
 */
#include <stdio.h>
#include <gc.h>
#include <stdarg.h>
#include <string.h>
#include "Str.h"
#include "myctype.h"

#define INITIAL_STR_SIZE 32

#ifdef STR_DEBUG
#define STR_LENGTH_CHECK(x) if (((x)->ptr==0&&(x)->length!=0)||(strlen((x)->ptr)!=(x)->length))abort();
#else
#define STR_LENGTH_CHECK(x)
#endif

Str
Strnew()
{
  Str             x = GC_MALLOC(sizeof(struct _Str));
  x->ptr = GC_MALLOC_ATOMIC(INITIAL_STR_SIZE);
  x->ptr[0] = '\0';
  x->area_size = INITIAL_STR_SIZE;
  x->length = 0;
  return x;
}

Str
Strnew_size(int n)
{
  Str             x = GC_MALLOC(sizeof(struct _Str));
  x->ptr = GC_MALLOC_ATOMIC(n + 1);
  x->ptr[0] = '\0';
  x->area_size = n + 1;
  x->length = 0;
  return x;
}

Str
Strnew_charp(char *p)
{
  Str             x = GC_MALLOC(sizeof(struct _Str));
  int             n;

  if (p == NULL)
    return Strnew();
  n = strlen(p) + 1;
  x->ptr = GC_MALLOC_ATOMIC(n);
  x->area_size = n;
  x->length = n - 1;
  strcpy(x->ptr, p);
  return x;
}

Str
Strnew_charp_n(char *p, int n)
{
  Str             x = GC_MALLOC(sizeof(struct _Str));
  int             len;

  if (p == NULL)
    return Strnew_size(n);
  len = strlen(p);
  if (len < n)
    n = len;
  x->ptr = GC_MALLOC_ATOMIC(n+1);
  x->area_size = n+1;
  x->length = n;
  strncpy(x->ptr, p, n);
  x->ptr[n] = '\0';
  return x;
}

Str
Strdup(Str s)
{
  Str             n = Strnew_size(s->length);
  STR_LENGTH_CHECK(s);
  Strcopy(n, s);
  return n;
}

void
Strclear(Str s)
{
  s->length = 0;
  s->ptr[0] = '\0';
}

void
Strfree(Str x)
{
  GC_free(x->ptr);
  GC_free(x);
}

void
Strcopy(Str x, Str y)
{
  STR_LENGTH_CHECK(x);
  STR_LENGTH_CHECK(y);
  if (x->area_size < y->length + 1) {
    GC_free(x->ptr);
    x->ptr = GC_MALLOC_ATOMIC(y->length + 1);
    x->area_size = y->length + 1;
  }
  strcpy(x->ptr, y->ptr);
  x->length = y->length;
}

void
Strcopy_charp(Str x, char *y)
{
  int             len;

  STR_LENGTH_CHECK(x);
  if (y == NULL) {
    x->length = 0;
    return;
  }
  len = strlen(y);
  if (x->area_size < len + 1) {
    GC_free(x->ptr);
    x->ptr = GC_MALLOC_ATOMIC(len + 1);
    x->area_size = len + 1;
  }
  strcpy(x->ptr, y);
  x->length = len;
}

void
Strcopy_charp_n(Str x, char *y, int n)
{
  int             len = n;

  STR_LENGTH_CHECK(x);
  if (y == NULL) {
    x->length = 0;
    return;
  }
  if (x->area_size < len + 1) {
    GC_free(x->ptr);
    x->ptr = GC_MALLOC_ATOMIC(len + 1);
    x->area_size = len + 1;
  }
  strncpy(x->ptr, y, n);
  x->ptr[n] = '\0';
  len = strlen(y);
  if (len < n)
    n = len;
  x->length = n;
}

void
Strcat_charp_n(Str x, char *y, int n)
{
  int newlen;

  STR_LENGTH_CHECK(x);
  if (y == NULL)
    return;
  newlen = x->length + n + 1;
  if (x->area_size < newlen) {
    char           *old = x->ptr;
    newlen = newlen * 3 / 2;
    x->ptr = GC_MALLOC_ATOMIC(newlen);
    x->area_size = newlen;
    bcopy(old, x->ptr, x->length);
    GC_free(old);
  }
  bcopy(y, &x->ptr[x->length], n);
  x->length += n;
  x->ptr[x->length] = '\0';
}

void
Strcat(Str x, Str y)
{
  STR_LENGTH_CHECK(y);
  Strcat_charp_n(x, y->ptr, y->length);
}

void
Strcat_charp(Str x, char *y)
{
  if (y == NULL) return;
  Strcat_charp_n(x, y, strlen(y));
}


void Strgrow(Str x)
{
  char           *old = x->ptr;
  int newlen;
  newlen = x->length * 6 / 5;
  if (newlen == x->length)
    newlen += 2;
  x->ptr = GC_MALLOC_ATOMIC(newlen);
  x->area_size = newlen;
  bcopy(old,x->ptr,x->length);
  GC_free(old);
}

Str
Strsubstr(Str s, int beg, int len)
{
  Str             new_s;
  int             i;

  STR_LENGTH_CHECK(s);
  new_s = Strnew();
  if (beg >= s->length)
    return new_s;
  for (i = 0; i < len && beg + i < s->length; i++)
    Strcat_char(new_s, s->ptr[beg+i]);
  return new_s;
}

void
Strlower(Str s)
{
  int             i;
  STR_LENGTH_CHECK(s);
  for (i = 0; i < s->length; i++)
    s->ptr[i] = tolower(s->ptr[i]);
}

void
Strupper(Str s)
{
  int             i;
  STR_LENGTH_CHECK(s);
  for (i = 0; i < s->length; i++)
    s->ptr[i] = toupper(s->ptr[i]);
}

void
Strchop(Str s)
{
  STR_LENGTH_CHECK(s);
  while ((s->ptr[s->length - 1] == '\n' || s->ptr[s->length - 1] == '\r') &&
	 s->length > 0) {
    s->length--;
  }
  s->ptr[s->length] = '\0';
}

void
Strshrink(Str s, int n)
{
  STR_LENGTH_CHECK(s);
  if (n >= s->length) {
    s->length = 0;
    s->ptr[0] = '\0';
  } else {
    s->length -= n;
    s->ptr[s->length] = '\0';
  }
}

void
Strshrinkfirst(Str s, int n)
{
  int i;
  STR_LENGTH_CHECK(s);
  if (n >= s->length) {
    s->length = 0;
    s->ptr[0] = '\0';
  } else {
    for (i = n; i <= s->length; i++)
      s->ptr[i-n] = s->ptr[i];
    s->length -= n;
  }
}
  

void
Strremovefirstspaces(Str s)
{
  int             i, j;

  STR_LENGTH_CHECK(s);
  for (i = 0; i < s->length && IS_SPACE(s->ptr[i]); i++);
  if (i == 0)
    return;
  s->length -= i;
  for (j = 0; j < s->length; j++)
    s->ptr[j] = s->ptr[j + i];
  s->ptr[j] = '\0';
}

void
Strremovetrailingspaces(Str s)
{
  int             i;

  STR_LENGTH_CHECK(s);
  for (i = s->length - 1; i >= 0 && IS_SPACE(s->ptr[i]); i--);
  s->length = i + 1;
  s->ptr[i + 1] = '\0';
}

Str
Stralign_left(Str s, int width)
{
  Str             n;
  int             i;

  STR_LENGTH_CHECK(s);
  if (s->length >= width)
    return Strdup(s);
  n = Strnew_size(width);
  Strcopy(n, s);
  for (i = s->length; i < width; i++)
    Strcat_char(n, ' ');
  return n;
}

Str
Stralign_right(Str s, int width)
{
  Str             n;
  int             i;

  STR_LENGTH_CHECK(s);
  if (s->length >= width)
    return Strdup(s);
  n = Strnew_size(width);
  for (i = s->length; i < width; i++)
    Strcat_char(n, ' ');
  Strcat(n, s);
  return n;
}

Str
Stralign_center(Str s, int width)
{
  Str             n;
  int             i, w;

  STR_LENGTH_CHECK(s);
  if (s->length >= width)
    return Strdup(s);
  n = Strnew_size(width);
  w = (width - s->length) / 2;
  for (i = 0; i < w; i++)
    Strcat_char(n, ' ');
  Strcat(n, s);
  for (i = w + s->length; i < width; i++)
    Strcat_char(n, ' ');
  return n;
}

#define SP_NORMAL 0
#define SP_PREC   1
#define SP_PREC2  2

Str
Sprintf(char *fmt,...)
{
  int             len = 0;
  int             status = SP_NORMAL;
  int             p;
  char           *f;
  Str             s;
  va_list         ap;

  va_start(ap,fmt);
  for (f = fmt; *f; f++) {
  redo:
    switch (status) {
    case SP_NORMAL:
      if (*f == '%') {
	status = SP_PREC;
	p = 0;
      } else
	len++;
      break;
    case SP_PREC:
      if (isalpha(*f)) {
	/* conversion char. */
	double vd;
	int vi;
	char *vs;
	void *vp;

	switch (*f) {
	case 'l':
	case 'h':
	case 'L':
	case 'w':
	  continue;
	case 'd':
	case 'i':
	case 'o':
	case 'x':
	case 'X':
	case 'u':
	  vi = va_arg(ap,int);
	  len += (p>0)?p:10;
	  break;
	case 'f':
	case 'g':
	case 'e':
	case 'G':
	case 'E':
	  vd = va_arg(ap,double);
	  len += (p>0)?p:15;
	  break;
	case 'c':
	  len += 1;
	  vi = va_arg(ap,int);
	  break;
	case 's':
	  vs = va_arg(ap,char*);
	  vi = strlen(vs);
	  len += (p>vi)?p:vi;
	  break;
	case 'p':
	  vp = va_arg(ap,void*);
	  len += 10;
	  break;
	case 'n':
	  vp = va_arg(ap,void*);
	  break;
	}
	status = SP_NORMAL;
      } else if (isdigit(*f))
	p = p * 10 + *f - '0';
      else if (*f == '.')
	status = SP_PREC2;
      else if (*f == '%') {
	status = SP_NORMAL;
	len++;
      }
      break;
    case SP_PREC2:
      if (isalpha(*f)) {
	status = SP_PREC;
	goto redo;
      }
      break;
    }
  }
  va_end(ap);
  s = Strnew_size(len * 2);
  va_start(ap, fmt);
  vsprintf(s->ptr, fmt, ap);
  va_end(ap);
  s->length = strlen(s->ptr);
  if (s->length > len*2) {
    fprintf(stderr,"Sprintf: string too long\n");
    exit(1);
  }
  return s;
}

Str
Strfgets(FILE *f)
{
  Str s = Strnew();
  char c;
  while (1) {
    c = fgetc(f);
    if (feof(f)) break;
    Strcat_char(s,c);
    if (c == '\n') break;
  }
  return s;
}

void
Strfputs(Str s, FILE *f)
{
    int i;
    for (i = 0; i < s->length; i++)
	putc(s->ptr[i],f);
}

Str
Strfgetall(FILE *f)
{
  Str s = Strnew();
  char c;
  while (1) {
    c = fgetc(f);
    if (feof(f)) break;
    Strcat_char(s,c);
  }
  return s;
}
