/* $Id: display.c,v 1.2 1999/07/29 11:50:45 okabe Exp $ */
#include <signal.h>
#include "fm.h"


#ifdef COLOR

#define EFFECT_ANCHOR_START       effect_anchor_start()
#define EFFECT_ANCHOR_END         effect_anchor_end()
#define EFFECT_IMAGE_START        effect_image_start()
#define EFFECT_IMAGE_END          effect_image_end()
#define EFFECT_FORM_START         effect_form_start()
#define EFFECT_FORM_END           effect_form_end()
#define EFFECT_ACTIVE_START	  effect_active_start()
#define EFFECT_ACTIVE_END	  effect_active_end()

/* color: 
 *  0  black
 *  1  red
 *  2  green
 *  3  yellow
 *  4  blue
 *  5  magenta
 *  6  cyan
 *  7  white
 */

#define EFFECT_ANCHOR_START_C       setfcolor(anchor_color)
#define EFFECT_IMAGE_START_C        setfcolor(image_color)
#define EFFECT_FORM_START_C         setfcolor(form_color)
#define EFFECT_ACTIVE_START_C	    setfcolor(active_color)

#define EFFECT_IMAGE_END_C          setfcolor(basic_color)
#define EFFECT_ANCHOR_END_C         setfcolor(basic_color)
#define EFFECT_FORM_END_C           setfcolor(basic_color)
#define EFFECT_ACTIVE_END_C	    setfcolor(basic_color)

#define EFFECT_ANCHOR_START_M       underline()
#define EFFECT_ANCHOR_END_M         underlineend()
#define EFFECT_IMAGE_START_M        standout()
#define EFFECT_IMAGE_END_M          standend()
#define EFFECT_FORM_START_M         standout()
#define EFFECT_FORM_END_M           standend()
#define EFFECT_ACTIVE_START_M	    underline()
#define EFFECT_ACTIVE_END_M	    underlineend()

#define define_effect(name_start,name_end,color_start,color_end,mono_start,mono_end) \
static void name_start { if (useColor) { color_start; } else { mono_start; }}\
static void name_end { if (useColor) { color_end; } else { mono_end; }}

define_effect(EFFECT_ANCHOR_START,EFFECT_ANCHOR_END,EFFECT_ANCHOR_START_C,EFFECT_ANCHOR_END_C,EFFECT_ANCHOR_START_M,EFFECT_ANCHOR_END_M)
define_effect(EFFECT_IMAGE_START,EFFECT_IMAGE_END,EFFECT_IMAGE_START_C,EFFECT_IMAGE_END_C,EFFECT_IMAGE_START_M,EFFECT_IMAGE_END_M)
define_effect(EFFECT_FORM_START,EFFECT_FORM_END,EFFECT_FORM_START_C,EFFECT_FORM_END_C,EFFECT_FORM_START_M,EFFECT_FORM_END_M)

static void EFFECT_ACTIVE_START
{
  if (useColor) {
    if (useActiveColor)
      EFFECT_ACTIVE_START_C;
    else {
      EFFECT_ANCHOR_START_C;
      EFFECT_ACTIVE_START_M;
    }
  }
  else
      EFFECT_ACTIVE_START_M;
}

static void EFFECT_ACTIVE_END
{
  if (useColor) {
    if (useActiveColor)
      EFFECT_ACTIVE_END_C;
    else {
      EFFECT_ANCHOR_END_C;
      EFFECT_ACTIVE_END_M;
    }
  }
  else
      EFFECT_ACTIVE_END_M;
}
    

#else /* no color */

#define EFFECT_ANCHOR_START       underline()
#define EFFECT_ANCHOR_END         underlineend()
#define EFFECT_IMAGE_START        standout()
#define EFFECT_IMAGE_END          standend()
#define EFFECT_FORM_START         standout()
#define EFFECT_FORM_END           standend()
#define EFFECT_ACTIVE_START       underline()
#define EFFECT_ACTIVE_END         underlineend()

#endif /* COLOR */


#ifndef JP_CHARSET
static char g_rule[] = "ntwluxkavmqajaaa";
#endif

/*
 * Terminate routine.
 */

void
fmTerm(void)
{
  Buffer *buf;
  char *f;

  if (fmInitialized) {
    move(LASTLINE, 0);
    clrtoeolx();
    refresh();
#ifdef MOUSE
    if (use_mouse)
      mouse_end();
#endif
    reset_tty();
    fmInitialized = NIL;
  }
}

void
deleteFiles()
{
  Buffer *buf;
  char *f;
  while (Firstbuf) {
    buf = Firstbuf->nextBuffer;
    discardBuffer(Firstbuf);
    Firstbuf = buf;
  }
  while ((f = popText(fileToDelete)) != NULL)
    unlink(f);
}


/*
 * Initialize routine.
 */
void
fmInit(void)
{
  if (!fmInitialized) {
    initscr();
    term_raw();
    term_noecho();
#ifdef MOUSE
    if (use_mouse)
      mouse_init();
#endif
  }
  fmInitialized = T;
}

/*
 * Display some lines.
 */
static Line    *cline = NULL;
static int      ccolumn = -1;

static int     ulmode = 0, somode = 0, bomode = 0;
static int     anch_mode = 0, emph_mode = 0, imag_mode = 0, form_mode = 0, active_mode = 0;
#ifndef JP_CHARSET
static int     graph_mode = 0;
#endif

void
displayBuffer(Buffer * buf, int mode)
{
  Str msg = Strnew();
  Anchor *aa;
  ParsedURL url;
  URLFile uf;

  if (buf->width == 0)
    buf->width = COLS;
  if (buf->height == 0)
    buf->height = LASTLINE+1;
  if (buf->width != COLS - 5 && buf->type && !strcmp(buf->type, "text/html"))
    reshapeBuffer(buf);
  if (mode == B_FORCE_REDRAW ||
      mode == B_SCROLL ||
      cline != buf->topLine ||
      ccolumn != buf->currentColumn) {
    if (mode == B_SCROLL && cline && buf->currentColumn == ccolumn) {
      int i, n = buf->topLine->linenumber -cline->linenumber;
      if (n > 0 && n < LASTLINE) {
	move(LASTLINE,0);
	clrtoeolx();
	refresh();
	for (i = 0; i < n; i++)
	  scroll();
      } else if (n < 0 && n > -LASTLINE) {
	for (i = 0; i < -n; i++)
	  rscroll();
      }
      redrawNLine(buf, n);
    } else
      redrawBuffer(buf);
    cline = buf->topLine;
    ccolumn = buf->currentColumn;
  }
  if (buf->firstLine == NULL) {
    clear();
    standout();
    msg = Sprintf("Viewing <%s>\tNo Line", cleanup_str(buf->buffername));
    message(msg->ptr, buf->cursorX, buf->cursorY);
    standend();
    refresh();
    return;
  }
  move(buf->cursorY, buf->cursorX);
  if (buf->topLine == NULL)
    buf->topLine = buf->firstLine;
  standout();
#ifdef MOUSE
  if (!use_mouse)
        msg = Sprintf("Viewing <%s> ", cleanup_str(buf->buffername));
  else
#if LANG == JA
  msg = Sprintf("㢬Viewing <%s> ", cleanup_str(buf->buffername));
#else
  msg = Sprintf("<=UpDn Viewing <%s> ", cleanup_str(buf->buffername));
#endif
#else
  msg = Sprintf("Viewing <%s> ", cleanup_str(buf->buffername));
#endif
  if (displayLink) {
    aa = retrieveCurrentAnchor(buf);
    if (aa) {
      ParsedURL url;
      Str s;
      int l;
      parseURL2(aa->url,&url,baseURL(buf));
      s = parsedURL2Str(&url);
      l = buf->width - msg->length;
      if(s->length > l && l >= 4) {
        Strcat(msg,Strsubstr(s, 0, (l - 2) / 2));
#if LANG == JA
        Strcat_charp(msg,"");
#else
        Strcat_charp(msg,"..");
#endif
        l = buf->width - msg->length;
        Strcat(msg,Strsubstr(s, s->length - l, l));
      } else {
	Strcat(msg,s);
      }
    }
  }
  message(msg->ptr, buf->cursorX, buf->cursorY);
  standend();
  refresh();
}

void
redrawBuffer(Buffer * buf)
{
  redrawNLine(buf, LASTLINE);
}

void
redrawNLine(Buffer * buf, int n)
{
  Line           *l, *l0;
  int             i;

#ifdef COLOR
  if (useColor) {
    EFFECT_ANCHOR_END_C;
#ifdef BG_COLOR
    setbcolor(bg_color);
#endif
  }
#endif
  for (i = 0, l = buf->topLine; i < LASTLINE; i++) {
    if (i >= LASTLINE-n || i < -n)
      l0 = redrawLine(buf, l, i);
    else {
      l0 = (l)? l->next : NULL;
    }
    if (l0 == NULL && l == NULL)
      break;
    l = l0;
  }
  if (n > 0)
    clrtobotx();
}

Line           *
redrawLine(Buffer * buf, Line * l, int i)
{
  int             j, pos, vpos, rcol;
  int             column = buf->currentColumn;
  char           *p;
  Lineprop       *pr;

  if (l == NULL) {
    if (buf->pagerSource) {
      l = getNextPage(buf, LASTLINE - i);
      if (l == NULL)
	return NULL;
    } else
      return NULL;
  }
  move(i, 0);
  if (l->len == 0 || COLPOS(l, l->len - 1) < column) {
    clrtoeolx();
    return l->next;
  }
  /* need_clrtoeol(); */
  pos = columnPos(l, column);
  p = &(l->lineBuf[pos]);
  pr = &(l->propBuf[pos]);
  vpos = COLPOS(l, pos);
  for (j = column; j < vpos; j++)
    addChar(' ', *pr);

  rcol = vpos;
#ifdef JP_CHARSET
  if (CharType(*pr) == PC_KANJI2) {
    addChar(' ', 0);
    rcol++;
  } else 
#endif
  if (*p == '\t') {
    do {
      addChar(' ', *pr);
      rcol++;
    } while ((rcol % Tabstop) != 0);
  } else {
    addChar(*p, *pr);
    rcol++;
  }

  if (column > vpos)
      column = vpos;
  for (j = 1; rcol - column < COLS && pos + j < l->len; j++) {
#ifdef JP_CHARSET
   if (rcol-column == COLS-1 && CharType(pr[j]) == PC_KANJI1)
      break;
#endif
   if (p[j] == '\t') {
      do {
	addChar(' ', 0);
      } while ((++rcol % Tabstop) != 0);
    } else {
      addChar(p[j], pr[j]);
      rcol++;
    }
  }
  if (somode) {
    somode = NIL;
    standend();
  }
  if (ulmode) {
    ulmode = NIL;
    underlineend();
  }
  if (bomode) {
    bomode = NIL;
    boldend();
  }
  if (emph_mode) {
    emph_mode = NIL;
    boldend();
  }

  if (anch_mode) {
    anch_mode = NIL;
    EFFECT_ANCHOR_END;
  }
  if (imag_mode) {
    imag_mode = NIL;
    EFFECT_IMAGE_END;
  }
  if (form_mode) {
    form_mode = NIL;
    EFFECT_FORM_END;
  }
  
  if (active_mode) {
    active_mode = NIL;
    EFFECT_ACTIVE_END;
  }
#ifndef JP_CHARSET
  if (graph_mode) {
    graph_mode = NIL;
    graphend();
  }
#endif
  if (rcol - column < COLS)
    clrtoeolx();
  return l->next;
}

int
redrawLineRegion(Buffer * buf, Line * l, int i, int bpos, int epos)
{
  int             j, pos, vpos, rcol;
  int             column = buf->currentColumn;
  char           *p;
  Lineprop       *pr;
  int            bcol, ecol;

  if (l == NULL)
    return 0;
  pos = columnPos(l, column);
  p = &(l->lineBuf[pos]);
  pr = &(l->propBuf[pos]);
  vpos = COLPOS(l, pos);

  rcol = vpos;
  bcol = bpos-pos;
  ecol = epos-pos;

  if (column > vpos)
    column = vpos;
  for (j = 0; rcol - column < COLS && pos + j < l->len; j++) {
#ifdef JP_CHARSET
    if (rcol-column == COLS-1 && CharType(pr[j]) == PC_KANJI1)
      break;
    if (rcol==column && CharType(pr[j]) == PC_KANJI2) {
      if (j >= bcol && j < ecol) {
	move(i, 0);
	addChar(' ', 0);
      }
      rcol++;
    } else 
#endif
    if (p[j] == '\t') {
      do {
	if (j >= bcol && j < ecol) {
	  move(i, rcol-column);
	  addChar(' ', 0);
	}
      } while ((++rcol % Tabstop) != 0);
    } else {
      if (j >= bcol && j < ecol) {
	move(i, rcol-column);
	addChar(p[j], pr[j]);
      }
      rcol++;
    }
  }
  if (somode) {
    somode = NIL;
    standend();
  }
  if (ulmode) {
    ulmode = NIL;
    underlineend();
  }
  if (bomode) {
    bomode = NIL;
    boldend();
  }
  if (emph_mode) {
    emph_mode = NIL;
    boldend();
  }

  if (anch_mode) {
    anch_mode = NIL;
    EFFECT_ANCHOR_END;
  }
  if (imag_mode) {
    imag_mode = NIL;
    EFFECT_IMAGE_END;
  }
  if (form_mode) {
    form_mode = NIL;
    EFFECT_FORM_END;
  }
  if (active_mode) {
    active_mode = NIL;
    EFFECT_ACTIVE_END;
  }
#ifndef JP_CHARSET
  if (graph_mode) {
    graph_mode = NIL;
    graphend();
  }
#endif
 
  return rcol - column;
}

#define do_effect1(effect,modeflag,action_start,action_end) \
if (m & effect) { \
	if (!modeflag) { \
		action_start; \
		modeflag = T; \
	} \
}

#define do_effect2(effect,modeflag,action_start,action_end) \
if (modeflag) { \
	action_end; \
	modeflag = NIL; \
}

void
do_effects(Lineprop m)
{
  /* effect end */
  do_effect2(PE_UNDER, ulmode, underline(), underlineend())
  do_effect2(PE_STAND, somode, standout(), standend())
  do_effect2(PE_BOLD, bomode, bold(), boldend())
  do_effect2(PE_EMPH, emph_mode, bold(), boldend())
  do_effect2(PE_ANCHOR, anch_mode, EFFECT_ANCHOR_START, EFFECT_ANCHOR_END);
  do_effect2(PE_IMAGE, imag_mode, EFFECT_IMAGE_START, EFFECT_IMAGE_END);
  do_effect2(PE_FORM, form_mode, EFFECT_FORM_START, EFFECT_FORM_END);
  do_effect2(PE_ACTIVE, active_mode, EFFECT_ACTIVE_START, EFFECT_ACTIVE_END);
#ifndef JP_CHARSET
  if (graph_mode) {
    graphend();
    graph_mode = NIL;
  }
#endif

  /* effect start */
  do_effect1(PE_UNDER, ulmode, underline(), underlineend())
  do_effect1(PE_STAND, somode, standout(), standend())
  do_effect1(PE_BOLD, bomode, bold(), boldend())
  do_effect1(PE_EMPH, emph_mode, bold(), boldend())
  do_effect1(PE_ANCHOR, anch_mode, EFFECT_ANCHOR_START, EFFECT_ANCHOR_END);
  do_effect1(PE_IMAGE, imag_mode, EFFECT_IMAGE_START, EFFECT_IMAGE_END);
  do_effect1(PE_FORM, form_mode, EFFECT_FORM_START, EFFECT_FORM_END);
  do_effect1(PE_ACTIVE, active_mode, EFFECT_ACTIVE_START, EFFECT_ACTIVE_END);
#ifndef JP_CHARSET
  if (m & PC_ISRULE) {
    if (!graph_mode && graph_ok()) {
      graphstart();
      graph_mode = T;
    }
  }
#endif
}

void
addChar(char c, Lineprop mode)
{
  Lineprop        m = CharEffect(mode);

#ifdef JP_CHARSET
  if (CharType(mode) != PC_KANJI2)
    do_effects(m);
  if (c & 0x80 && CharType(mode) == PC_ASCII) {
    addch(' ');
  } else 
#else
  do_effects(m);
  if (graph_mode)
    addch(g_rule[PC_RULE_VAL(m)]);
  else
#endif
  if (c == '\t') {
    addch(c);
  } else if (!(c & ~0x1f) && c != CTRL_J) {	/* Control code */
    addch('^');
    addch(c + '@');
  } else if (c == DEL_CODE)
    addstr("^?");
  else if (c != '\n')
    addch(c);
  /* do nothing for \n */
}

void
message(char *s, int return_x, int return_y)
{
  if (!fmInitialized)
    return;
  move(LASTLINE, 0);
  addnstr(s, COLS - 1);
  clrtoeolx();
  move(return_y, return_x);
}

void
disp_message_nsec(char *s, int redraw_current, int sec, int purge)
{
  if (!fmInitialized) {
    fprintf(stderr,"%s\n",s);
    return;
  }
  if (Currentbuf != NULL)
    message(s,Currentbuf->cursorX,Currentbuf->cursorY);
  else
    message(s,LASTLINE,0);
  refresh();
  sleep_till_anykey(sec, purge);
  if (Currentbuf != NULL && redraw_current)
    displayBuffer(Currentbuf,B_NORMAL);
}

void
disp_message(char *s, int redraw_current)
{
  disp_message_nsec(s, redraw_current, 10, 0);
}

void
cursorUp(Buffer * buf)
{
  if (buf->firstLine == NULL)
    return;
  if (buf->cursorY > 0)
    cursorUpDown(buf, -1);
  else {
    buf->topLine = lineSkip(buf->topLine, -(LASTLINE+1) / 2);
    if (buf->currentLine->prev != NULL)
      buf->currentLine = buf->currentLine->prev;
    buf->cursorY = buf->currentLine->linenumber - buf->topLine->linenumber;
  }
}

void
cursorDown(Buffer * buf)
{
  if (buf->firstLine == NULL)
    return;
  if (buf->cursorY < LASTLINE - 1)
    cursorUpDown(buf, 1);
  else {
    buf->topLine = lineSkip(buf->topLine, (LASTLINE+1) / 2);
    if (buf->currentLine->next != NULL)
      buf->currentLine = buf->currentLine->next;
    buf->cursorY = buf->currentLine->linenumber - buf->topLine->linenumber;
  }
}

void
cursorUpDown(Buffer * buf, int n)
{
  int             i;
  Line           *cl = buf->currentLine;

  if (buf->firstLine == NULL)
    return;
  if ((buf->currentLine = lineSkip(cl, n)) == cl)
    return;
  buf->cursorY = buf->currentLine->linenumber - buf->topLine->linenumber;
  i = columnPos(buf->currentLine, buf->currentColumn + buf->visualpos);
  buf->cursorX = COLPOS(buf->currentLine, i) - buf->currentColumn;
  buf->pos = i;
  if (buf->cursorX < 0)
    buf->cursorX = 0;
}

void
cursorRight(Buffer * buf)
{
  int             i, delta = 1;
  Line           *l = buf->currentLine;
  Lineprop       *p;

  if (buf->firstLine == NULL)
    return;
  i = buf->pos;
  p = l->propBuf;
#ifdef JP_CHARSET
  if (CharType(p[i]) == PC_KANJI1)
    delta = 2;
#endif
  if (i >= buf->currentLine->len - delta)
    return;
  if (buf->visualpos >= COLS - delta)
    columnSkip(buf, COLS / 2);
  buf->visualpos = COLPOS(l, i + delta) - buf->currentColumn;
  buf->cursorX = buf->visualpos;
  buf->pos += delta;
}

void
cursorLeft(Buffer * buf)
{
  int             i, delta = 1;
  Line           *l = buf->currentLine;
  Lineprop       *p;

  if (buf->firstLine == NULL)
    return;
  i = buf->pos;
  if (i <= 0)
    return;
  p = l->propBuf;
#ifdef JP_CHARSET
  if (CharType(p[i - 1]) == PC_KANJI2)
    delta = 2;
#endif
  if (buf->visualpos <= delta - 1)
    columnSkip(buf, -COLS / 2);
  buf->visualpos = COLPOS(l, i - delta) - buf->currentColumn;
  buf->cursorX = buf->visualpos;
  buf->pos = i - delta;
}

void
cursorHome(Buffer * buf)
{
  buf->visualpos = 0;
  buf->cursorX = buf->cursorY = 0;
}


/*
 * Arrange line,column and cursor position according to current line and
 * current position.
 */
void
arrangeCursor(Buffer * buf)
{
  int             col;
  if (buf == NULL || buf->currentLine == NULL)
    return;
  /* Arrange line */
  if (buf->currentLine->linenumber - buf->topLine->linenumber >= LASTLINE ||
      buf->currentLine->linenumber < buf->topLine->linenumber) {
    buf->topLine = buf->currentLine;
  }
  /* Arrange column */
  col = COLPOS(buf->currentLine, buf->pos);
  if (col < buf->currentColumn || col > buf->currentColumn + COLS) {
    buf->currentColumn = 0;
    if (col >= COLS)
      columnSkip(buf, col);
  }
  /* Arrange cursor */
  buf->cursorY = buf->currentLine->linenumber - buf->topLine->linenumber;
  buf->visualpos = buf->cursorX = COLPOS(buf->currentLine, buf->pos) - buf->currentColumn;

}

void
cursorXY(Buffer * buf, int x, int y)
{
  int oldX;

  cursorUpDown(buf, y - buf->cursorY);

  if (buf->cursorX > x) {
    while (buf->cursorX > x)
      cursorLeft(buf);
  } else if (buf->cursorX < x) {
    while (buf->cursorX < x) {
      oldX = buf->cursorX;

      cursorRight(buf);

      if (oldX == buf->cursorX)
        break;
    }
    if (buf->cursorX > x)
      cursorLeft(buf);
  }
}

